/* 
 * Copyright (c)2007-2008, Dustin Spicuzza
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * The name of the Dustin Spicuzza may not be used to endorse or promote 
 *       products derived from this software without specific prior written 
 *       permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Dustin Spicuzza ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Dustin Spicuzza BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using System;
using System.Windows.Forms;
using System.Drawing;
using System.ComponentModel;

namespace Panes {

	/// <summary>
	/// The base to inherit from. Implements a lot of common functionality, since the
	/// panes and the panecontainers are pretty much the same, except for a few key points. 
	/// </summary>
	public partial class PaneBase : ContainerControl {

		/// <summary>
		/// Debugging ID
		/// </summary>
		internal int aId = 0;

		internal PositionControls positionControls = null;
		internal bool canDisplayDragEdges = true;

		private bool isRoot = true;
		private Splitter _mySplitter = null;

		/// <summary>
		/// This event is called when the pane is closed by the user.
		/// </summary>
		public event EventHandler PaneClosing = null;

		bool isClosed = false;

		/// <summary>
		/// Set to true if the pane has been closed
		/// </summary>
		public bool IsClosed {
			get { return isClosed; }
		}


		/// <summary>
		/// If there is a splitter associated with this, then it will be set to non-null
		/// </summary>
		internal Splitter mySplitter {
			get { return _mySplitter; }
			set { 
				_mySplitter = value;
				// helps w/ debugging
				if ( _mySplitter != null ) {
					_mySplitter.Name = "Split" + this.Name;
					//_mySplitter.Height = _mySplitter.Width = (int)Math.Ceiling( (float)Font.Height / 3 );
				}
			}
		}

		public PaneBase() {
			Panes.Add(this);
			this.aId = PaneBase.saId++;
			this.Name = "Panebase" + aId.ToString();
		}


		protected override void OnCreateControl() {
			base.OnCreateControl();

			// don't add controls in design mode, since its pretty annoying to worry about
			if ( !this.DesignMode ) {

				positionControls = new PositionControls(this);
				
				// these properties rely on the control being created
				IsRoot = isRoot;
			}
		}

		protected override void OnControlAdded(ControlEventArgs e) {
			base.OnControlAdded(e);

			// this is far too annoying to implement correctly, screw it
			if ( this.DesignMode )
				throw new Exception("You cannot add components to this using the designer!");
		}

		protected override void Dispose(bool disposing) {

			if ( !IsClosed ) {
				Close();
				isClosed = true;
			}

			if ( mySplitter != null ) {
				mySplitter.Dispose();
				mySplitter = null;
			}

			if ( RootPanes.Contains(this) )
				RootPanes.Remove(this);

			if ( Panes.Contains(this) )
				Panes.Remove(this);

			positionControls = null;

			base.Dispose(disposing);
		}


		protected override void OnParentChanged(EventArgs e) {
			base.OnParentChanged(e);

			if ( this.Parent == null )
				PaneBase.Panes.Remove(this);
			else if ( !PaneBase.Panes.Contains(this) )
				PaneBase.Panes.Add(this);

		}


		/// <summary>
		/// Set this to true if this is the root PaneContainer
		/// </summary>
		[DefaultValue(true)]
		public virtual bool IsRoot {
			get { return isRoot; }
			set {
				isRoot = value;

				if ( value && !PaneBase.RootPanes.Contains(this) )
					PaneBase.RootPanes.Add(this);
				else if ( !value && PaneBase.RootPanes.Contains(this) )
					PaneBase.RootPanes.Remove(this);
			}
		}

		/// <summary>
		/// This function is called during a mouse mouse of a floating pane, to determine whether or not
		/// to show the drag edges or not. This will try to call all of its children as well. 
		/// </summary>
		/// <param name="mousePosition">Point to check</param>
		/// <param name="hasMouse">This is set to true if you might have the mouse positioned inside of you. If it
		/// is set to false, then theres no way the mouse is inside of you.</param>
		/// <returns>Returns either tryMouse, or false if it displayed a drag window</returns>
		internal virtual bool UpdateDragWindow(Point mousePosition, bool tryMouse) {

			if (!canDisplayDragEdges)
				return tryMouse;

			// check yourself first
			if ( tryMouse) {

				if ( RectangleToScreen(ClientRectangle).Contains(mousePosition) ) {

					positionControls.UpdateDragWindows(mousePosition);
					return false;
				}
			}

			positionControls.HideControls();

			return tryMouse;
		}

		/// <summary>
		/// Default implementation of addPane
		/// </summary>
		/// <param name="p"></param>
		/// <param name="position"></param>
		public virtual void AddPane(PaneBase pb, DockStyle position) {

			Control parent = Parent;
			PaneBase npb = null;

			parent.SuspendLayout();
			this.SuspendLayout();

			// replace self with either a tabpane or a panecontainer
			if ( position == DockStyle.Fill ) {
				npb = new TabPane(TabPane.TabPaneStyle.Bottom);
			} else {
				npb = new PaneContainer(false);
			}

			// set parent properties
			if ( this.Parent.GetType() == typeof(FloatingPane) ){
				this.Parent.Text = "";
				npb.IsRoot = true;
			} else {
				npb.mySplitter = mySplitter;
				mySplitter = null;
			}

			this.Parent.Controls.Add(npb);
			this.Parent.Controls.SetChildIndex(npb, this.Parent.Controls.GetChildIndex(this));
			npb.Dock = this.Dock;
			npb.Size = this.Size;
				
			// remove self from parent
			this.Parent = null;
			this.IsRoot = false;

			pb.Parent = null;

			// add self, splitter, next
			if ( position != DockStyle.Fill ) {

				Splitter s = new Splitter();

				npb.Controls.Add(this);
				this.Dock = DockStyle.Fill;		// TODO: Is this something that should be done?

				// associate the new splitter with the incoming control
				s.Dock = pb.Dock = position;
				pb.mySplitter = s;

				npb.Controls.Add(s);
				npb.Controls.Add(pb);

				// finally, ensure each component has equal size, if we're in a paneContainer
				if ( position == DockStyle.Left || position == DockStyle.Right )
					pb.Width = this.Width = npb.Width / 2;
				else
					pb.Height = this.Height = npb.Height / 2;

			} else {

				// add self, then add the item
				npb.AddPane(this, DockStyle.Fill);
				npb.AddPane(pb, DockStyle.Fill);
			}

			this.ResumeLayout();
			parent.ResumeLayout();
			pb.IsRoot = false;
		}

		public virtual void Close() {

			// this should be overridden, and the base function (this one) called. 

			// launch this event first
			if ( PaneClosing != null )
				PaneClosing(this, EventArgs.Empty);

			isClosed = true;
		}
	}
}
